#!/usr/bin/env python3

"""
Simple web server for controlling a remote db6nt
Oscar 100 upconvter

supports request of mode
tx forward / tx reflected / tx on/off
power on / off
setting power on/off

(c) dg7bbp 2022
"""

import argparse
import os
import morepath
import waitress
import serial
import collections
import time
import threading
import random
import json
import math

from static_resources import ResourcesApp, get_fileresource
from webfiles import web_dir
from antenna_switch import AntennaSwitch

_version = 1.08
_verbose = 0
_demo = 0

def log(level, msg):
   if level <= _verbose:
       print(msg) 

def logserial(direction, msg):
    if _verbose == 5:
       with open("mk_log.txt", "a+") as f:
          f.write("%s : '%s'\n" % (direction, msg))

class MkuControl(object):
    def __init__(self,portdevice):
        self.portdevice = portdevice
        self._connection = None
        self.cmd_queue = collections.deque() 
        self.loop = False
        self.thread = None 
        self.status = {"error": "No Signal"}
        self.swr_alarm = False
        self._standby = False
        self._gps_status = False

    def __enter__(self):
       self.connect()
       return self

    def __exit__(self, exc_type, exc_value, traceback): 
       self.disconnect()

    def connect(self):
      if self._connection is not None:
          self.disconnect()
      if not _demo:
          self._connection = serial.Serial(self.portdevice, 115200, timeout=0.1, write_timeout=0.5)
      self.thread = threading.Thread(target=self.status_loop, args=())
      self.loop = True
      self.thread.start() 

    def disconnect(self):
        if self.loop and self.thread:
            self.loop = False
            self.thread.join()
        if self._connection is not None:
            self._connection.close()
            self._connection = None
    
    def send_cmd(self, cmd, waittime=0.1):
      if self._connection is None:
      	return None
      unexpected = self._connection.read(100)
      logserial("unexpected <", str(unexpected))
      if unexpected:
          log(1, "Unexpected: :'%s'" % str(unexpected))
      log(3, "CMD: %s" % cmd)
      no_bytes = self._connection.write("{}\r".format(cmd).encode("utf-8"))
      logserial("> ", "cmd: '%s' + Bytes(%s)" % (cmd, no_bytes))
      self._connection.flush()
      # read until no \r oder no char is detected
      reply = ""
      time.sleep(waittime)
      bchar = self._connection.read(1)
      log(3, "CHAR: '%s' " % str(bchar))
      while bchar  and bchar !=  b"\r":
          reply += bchar.decode("utf-8")
          bchar = self._connection.read(1)
          log(3, "CHAR : '%s' %s '%s'" % (str(bchar), ord(bchar.decode("utf-8")) if bchar else "-", reply))
      if not bchar:
          reply = None
      logserial("data <", reply)
      log(2, "Reply: %s"% reply)
      return reply

    def power(self, on):
        """
        """ 
        self.cmd_queue.append((self._runpower, {"on": on})) 

    def _runpower(self, on=False):
       if on:
          reply = self.send_cmd("O1", 5)
       else:
          reply = self.send_cmd("O0", 5)
       return self._r_error(reply)

    def delete_swr(self):
       self.cmd_queue.append((self._delete_swr, {}))

    def _delete_swr(self):
       self.swr_alarm = False
       reply = self.send_cmd("S")
       return self._r_error(reply)

    def _r_error(self, reply, check_int=False):
        ret =  reply is None or reply.startswith("*")
        if not ret and check_int:
           if not reply.isdigit():
              log(0, "invalid integer: '%s' " % ret)
              ret = True
        return ret
 
    def status_loop(self):
        first = True
        while self.loop:
             self.run_status(first)
             first = False

    def get_status(self):
        if _demo:
            time.sleep(0.5)
            demo_name = os.path.join(web_dir(),
                                     "demo",
                                     "demo%02d.json" % random.randrange(_demo))
            if os.path.isfile(demo_name):
                with open(demo_name,"rb") as df:
                   return json.load(df)
        return self.status

    def standby(self, s):
        self._standby = s

    def run_status(self, with_gps):
      """
      """
      global _demo
      if _demo:
          time.sleep(2)
          return
      while len(self.cmd_queue):
          c = self.cmd_queue.pop()
          log(3, "running command")
          c[0](**c[1])

      fwd_int = 0
      refl_int = 0
      ptt_bool = False
      power_bool = False
      swr_alarm_bool = False
      gps_bool = False
      temp_int = 0
      swr = 0
      fwd_pwr = 0
      if self._standby:
          time.sleep(1)
      else:
          power = self.send_cmd("o")
          if not self._r_error(power):
              power_bool = power == "1"
          if True: 
             ptt = self.send_cmd("p")
             if not self._r_error(ptt):
                 ptt_bool = ptt == "1"
          if True: 
              fwd = self.send_cmd("f")
              refl = self.send_cmd("r")
              if not self._r_error(fwd, True):
                  fwd_int = int(fwd)
              if not self._r_error(refl, True):
                  refl_int = int(refl)      
          if with_gps: 
              gps = self.send_cmd("l")
              if not self._r_error(gps):
                  gps_bool = gps == "1"
                  self._gps_status = gps_bool
          else:
              gps_bool = self._gps_status

          temp = self.send_cmd("t")
          if not self._r_error(temp):
             temp_int = int(temp)
          swr_alarm = self.send_cmd("s")
          if not self._r_error(swr_alarm):
              swr_alarm_bool = swr_alarm == "1"
          # fwd_int is Voltage from ADC
          # 20 = U*U/50.0
          # umax = sqrt(20*50)
          p_max = 40.0
          u_max = math.sqrt(p_max*50)
          fwd_pwr = (fwd_int * u_max / 255.0)**2  / 50.0
          #fwd_pwr = 20.0 *fwd_int / 255
          swr = 1.0
          if fwd_int > 0:
              ref = float(refl_int)/float(fwd_int)
              if ref >= 1.0:
                  ref = 0.999
              swr = (1.0 + ref) / (1.0 - ref) 
              log(1, "swr: %s" % swr)
          if not ptt_bool:
              swr = 0
              fwd_int = 0
              fwd_power = 0
      self.status = {"forward": fwd_int,
                     "reflected": refl_int,
                     "power": power_bool,
                     "ptt": ptt_bool,
                     "temperature": temp_int,
                     "swr": swr,
                     "fwd_power": fwd_pwr,
                     "gps": gps_bool,
                     "swr_alarm": swr_alarm_bool}

  

class RootApp(morepath.App):
    pass

_mku_connection = None 

class MkuModel(object):
    def __init__(self, cmd):
        self.cmd = cmd

    def execute_cmd(self):
        global _mku_connection
        if self.cmd == "poweron":
            _mku_connection.power(True)
            ret =  _mku_connection.get_status()
        elif self.cmd == "poweroff":
            _mku_connection.power(False)
            ret =  _mku_connection.get_status()
        elif self.cmd == "resetswr":
            _mku_connection.delete_swr()
            ret =  _mku_connection.get_status()
        elif self.cmd == "status":
            ret =  _mku_connection.get_status()
        elif self.cmd == "standby":
            ret = _mku_connection.standby(True)
        elif self.cmd == "resume":
            ret = _mku_connection.standby(False)
        else:
            ret = ["unkown command"]
        return ret
                 
_ant_switch = None

class AntennaModel(object):
   def __init__(self, antenna):
       self.antenna = None
       if antenna is not None and antenna.isdigit():
           self.antenna = int(antenna)              

   
   def set_antenna(self):
       global _ant_switch
       if _ant_switch is not None:
          if self.antenna is not None:
              if self.antenna >= 0 and self.antenna <=7:
                 _ant_switch.on()
                 _ant_switch.set_antenna(self.antenna)
              elif self.antenna == 9:
                  _ant_switch.off()
       return [self.antenna]

@RootApp.path(path='')
class Root(object):
    pass

@RootApp.path(model=MkuModel, path='mku/{cmd}' )
def run_cmd(cmd):
    return MkuModel(cmd)

@RootApp.path(model=AntennaModel, path='ant/{antenna}')
def set_ant(antenna):
    return AntennaModel(antenna)

@RootApp.json(model=AntennaModel)
def antenna_set(self, request):
    return self.set_antenna()

@RootApp.html(model=Root)
def hello_word(self, request):
     global _ant_switch
     html = ""
     fname =  "indexant.html" if _ant_switch else "index.html"
     with open(os.path.join(web_dir(), "web", fname), "rb") as f:
          html = f.read()
     return html

@RootApp.json(model=MkuModel) 
def mku_cmd(self, request):
    return self.execute_cmd()

@RootApp.mount(app=ResourcesApp, path="web")
def mount_static_resources():
    return ResourcesApp()


def run_wsgi_server():
    global _mku_connection
    global _ant_switch
    global _verbose
    global _demo
    ap = argparse.ArgumentParser()
    ap.add_argument("--serial", required=True, help="Serial port of mku 2424")
    ap.add_argument("--port", type=int, default=8080, help="Web service listen port (default: 8080)")
    ap.add_argument("--verbose", type=int, default=0, help="Verbose level 0..4")
    ap.add_argument("--antennaswitch", action="store_true", help="enabled antennaswitch")
    ap.add_argument("--demo", type=int, default=0, help="Use demofiles")
    args = ap.parse_args()
    _verbose = args.verbose
    _demo = args.demo
    _antennaswitch = args.antennaswitch
    with MkuControl(args.serial) as _mku_connection:  
       _mku_connection.connect()
       if args.antennaswitch:
           with AntennaSwitch() as _ant_switch:
               _ant_switch.on()
               waitress.serve(RootApp(),
                              host="*",
                              port=args.port)
       else:
           _ant_switch = None
           waitress.serve(RootApp(),
                          host="*",
                          port=args.port)


if __name__ == '__main__':
    #morepath.run(RootApp())
    run_wsgi_server()
